/** 
 *  This file contains all necessary functions and data structure to operate with lifeCycles object
 *  defined via JSON format. For more details about certain format see: 
 *  http://asteasolutions.net/mediawiki/index.php/Sophie_JSON#LifeCycle
 */
  
var bookLifeCycle;// = new Array();
var pageLifeCycle;// = new Array();
/*dictionary for every object is stored, every element of it contains array with life cycle objects
 * lifeCycles[idOfTheFrame][introLength] -> integer
 * 						   [mainLength] -> integer
 * 						   [outroLength] -> integer
 *                         [looping] -> boolean
 * 						   [initialIntro] -> Array from objects {target, property, value}
 * 						   [initialMain]
 * 						   [initialOutro]
 * 						   [intro] -> Array from objcts {target, property, value, time}
 * 						   [main]
 * 						   [outro]
*/
var groupLifeCycles;// = new Array();
//dictionary for every object, every element of it contains array with life cycle objects
var lifeCycles;// = new Object();
var bookTimer;
var pageTimer;
// store timers for the current page
var frameTimers;

// store paused position of book audio for every page
var pausedAudioPositions;
// store initial position of book audio for every page
var bookAudioPositions;

// Booleans that determine whether after finishing outro should go to next or previous page of the book
var goToNextPageAfterOutro = false;
var goToPreviousPageAfterOutro = false;

// Constants for phases names
const INTRO = "INTRO";
const MAIN = "MAIN";
const OUTRO = "OUTRO";
const OFF = "OFF";

// Constants with names of the arrays with initial values of properties of each of the phases
const INITIAL_INTRO = "initialIntro";
const INITIAL_MAIN = "initialMain";
const INITIAL_OUTRO = "initialOutro";

// Constants with names of the fields of the array with life cycle that contains lengths of the phases in milliseconds
const INTRO_LENGTH = "introLength";
const MAIN_LENGTH = "mainLength";
const OUTRO_LENGTH = "outroLength";

// Constants for properties that can be changed through life cycle
const VISIBILITY = "visibility";
const LOCATION = "location";
const MEDIA_STATE = "mediastate";
const ACTIVATION_FOR_FRAME = "activationfor";

/**
 * Sort function used to store life cycles sorted
 * @param {Object} a
 * @param {Object} b
 */
function sortByTime(a, b){
	//Compare "a" and "b" in some fashion, and return -1, 0, or 1
	return a.time - b.time;
}

/**
 * Function for getting name of the previous phase of the life cycle
 * @param {String} currentPart - current phase of the life cycle
 */
function getPreviousArrayName(currentPart)
{
	switch (currentPart)
	{
		case INTRO:
			return null;
		case MAIN:
			return INTRO;
		case OUTRO:
			return MAIN;
		default:
			return OFF;
	}
}

/**
 * Function for getting name of the next phase of the life cycle
 * @param {String} currentPart - current phase of the life cycle
 */
function getNextArrayName(currentPart)
{
	switch (currentPart)
	{
		case INTRO:
			return MAIN;
		case MAIN:
			return OUTRO;
		case OUTRO:
			return OFF;
		default:
			return null;
	}
}

/**
 * Function for getting name of the array with initial values of properties for some of the phases
 * @param {String} currentPart - current phase of the life cycle
 */
function getInitialArrayName(currentPart)
{
	switch (currentPart)
	{
		case INTRO:
			return INITIAL_INTRO;
		case MAIN:
			return INITIAL_MAIN;
		case OUTRO:
			return INITIAL_OUTRO;
		default:
			return null;
	}
}

/**
 * Function for getting of the length of the phase of life cycle
 * @param {long} lifeCycle - current life cycle
 * @param {long} part - current phase of the life cycle
 */
function getLengthOfPart(lifeCycle, part)
{
	switch (part)
	{
		case INTRO:
			return lifeCycle[INTRO_LENGTH];
		case MAIN:
			return lifeCycle[MAIN_LENGTH];
		case OUTRO:
			return lifeCycle[OUTRO_LENGTH];
		default:
			return -1;
	}
}

/**
 * Creates life cycles for the sophie frame
 * @param {Object} data - json of the sophie frame
 */  
function buildFrameLifeCycle(data, pageId)
{
	if (lifeCycles == null)
		lifeCycles = new Array();
	if (lifeCycles[pageId] == null)
		lifeCycles[pageId] = new Array();
		
	// building array object for the life cycle of the frame 
	if (lifeCycles[pageId][data.id] == null)
		lifeCycles[pageId][data.id] = new Array();
	else
		return;
	
	buildLifeCycle(lifeCycles[pageId][data.id], data);

	// getting initial properties of the parts of the life cycle
	getAndStoreInitialProperties("frame", data.id, lifeCycles[pageId][data.id]);
	
	// creating LifeCycle object for the life cycle
	if (frameTimers == null)
		frameTimers = new Array();// Object();
	if (frameTimers[pageId] == null)
		frameTimers[pageId] = new Object();
	frameTimers[pageId][data.id] = new LifeCycle(data.id, "frameLifeCycle", lifeCycles[pageId][data.id], 0);
}

/**
 * Creates life cycles for the sophie page
 * @param {Object} data - json of the sophie page
 */
function buildPageLifeCycle(data)
{
	if (pageLifeCycle == null)
		pageLifeCycle = new Array();
	if (pageLifeCycle[data.id] == null)
		pageLifeCycle[data.id] = new Array();
	else
		return;
	
	if(bookAudioPositions == null)
		bookAudioPositions = new Array();
		
	if(pausedAudioPositions == null)
		pausedAudioPositions = new Array();
			
	//set audio position
	if(data.audioPosition != null)
	{
		bookAudioPositions[data.id] = data.audioPosition;
		//console.log("bookAudioPositions[" + data.id + "] is set to " + bookAudioPositions[data.id]);
	}
		
	// building life cycle of the page and storing it into pageLifeCycle
	buildLifeCycle(pageLifeCycle[data.id], data);
	
	//var framesAtThePage = document.getElementById(data.id).children;
	// getting initial positions of life cycles of the each frame over the page for main and outro of the page life cycle
	if (pageLifeCycle[data.id][INITIAL_INTRO] != null)
		$.each(pageLifeCycle[data.id][INITIAL_INTRO], function(index, item) 
		{ 
			if (item.property != VISIBILITY)
				getAndStoreInitialProperties("page", item.target, pageLifeCycle[data.id]);
		});
	/*for (var i=1; i < framesAtThePage.length; i++)
	{
		getAndStoreInitialProperties("page", framesAtThePage[i], pageLifeCycle[data.id]);
	}*/
	
	if (pageTimer == null)
		pageTimer = new Array();
	if (pageTimer[data.id] == null)
		pageTimer[data.id] = new Object();
	// creating LifeCycle object for life cycle of the page
	pageTimer[data.id] = new LifeCycle(data.id, "pageLifeCycle", pageLifeCycle[data.id], 0);
}

/**
 * Creates life cycles for the sophie book
 * @param {Object} data - json of the sophie book
 */
function buildBookLifeCycle(data)
{
	// building life cycle of the book and storing in into bookLifeCycle
	buildLifeCycle(bookLifeCycle, data);
	
	// getting initial page numbers for main and outro parts of the book life cycle	
	getMainInitialProperty(bookLifeCycle, "book", "pageNumber");
	getOutroInitialProperty(bookLifeCycle, "book", "pageNumber");
	
	// creating LifeCycle object for life cycle of the book
	bookTimer = new LifeCycle("bookLifeCycle", bookLifeCycle, 0);
	//starting book life cycle
	bookTimer.start();

}

/**
 * Creates life cycles for the sophie group
 * @param {Object} data - json of the sophie group
 */
function buildGroupLifeCycle(data)
{
	if (groupLifeCycles[data.id] == null)
		groupLifeCycles[data.id] = new Array();
	if(data.lifeCycle != null)
		buildLifeCycle(groupLifeCycles[data.id], data);
	
	// creating LifeCycle object for the life cycle
	if (frameTimers == null)
		frameTimers = new Object();
	frameTimers[data.id] = new LifeCycle(data.id, groupLifeCycles[data.id], 0);
	frameTimers[data.id].start();
}

/**
 * Creates life cycle from json
 * @param {Array} lifeCycle - array for storing the life cycle
 * @param {Object} data - json of the object for life cycle
 */
function buildLifeCycle(lifeCycle, data)
{
	var i;
	var intro = parseInt(data.lifeCycle.intro);
	var main = parseInt(data.lifeCycle.main);
	var outro = parseInt(data.lifeCycle.outro);
	var currentElementTime = 0;
	// storing lengths of the parts of the life cycle
	lifeCycle[INTRO_LENGTH] = intro;
	lifeCycle[MAIN_LENGTH] = main;
	lifeCycle[OUTRO_LENGTH] = outro;
	lifeCycle["looping"] = data.lifeCycle.looping;
	
	for (i = 0; i < data.lifeCycle.cycles.length; i++) 
	{
		// initial properties are set immediatelly and stored as initial properties for intro of the life cycle
		if (data.lifeCycle.cycles[i].time.milliseconds == 0)
		{
			var phaseArrayIndex = getInitialArrayName(data.lifeCycle.cycles[i].time.phase);
			if (phaseArrayIndex != null) {
				if (lifeCycle[phaseArrayIndex] == null) 
					lifeCycle[phaseArrayIndex] = new Array();
				lifeCycle[phaseArrayIndex].push({
					"target": data.lifeCycle.cycles[i].target,
					"property": data.lifeCycle.cycles[i].property,
					"value": data.lifeCycle.cycles[i].value
				});
			}
		}
		else 
		{
			currentElementTime = parseInt(data.lifeCycle.cycles[i].time.milliseconds);
			// gets cycle's position over the life cycle (intro, main or oustro)
			var positionInLifeCycle = data.lifeCycle.cycles[i].time.phase;
			if(lifeCycle[positionInLifeCycle] == null)
					lifeCycle[positionInLifeCycle] = new Array();
			
			// storing cycle in to the array of the necessary part of the life cycle
			lifeCycle[positionInLifeCycle].push({
				"target": data.lifeCycle.cycles[i].target,
				"property": data.lifeCycle.cycles[i].property,
				"value": data.lifeCycle.cycles[i].value,
				"time": data.lifeCycle.cycles[i].time.milliseconds
			});
		} 
	}
	
	//sort lifeCycles of every object
	if (lifeCycle[INTRO] != null)
		lifeCycle[INTRO].sort(sortByTime);
	if (lifeCycle[MAIN] != null)
		lifeCycle[MAIN].sort(sortByTime);
	if (lifeCycle[OUTRO] != null)
		lifeCycle[OUTRO].sort(sortByTime);
}
/**
 * Finds and stores initial values for all of the properties for life cycle
 * @param {String} dataType - type of the life cycle ("page" or "frame")
 * @param {String} dataId - id of the element which is the life cycle
 * @param {Array} lifeCycle - life cycle array
 */
function getAndStoreInitialProperties(dataType, dataId, lifeCycle)
{
	var value;
	var properties = getProperties(dataType);
	for (var i=0; i < properties.length; i++)
	{
		value = getInitialProperty(lifeCycle, MAIN, dataId, properties[i]);
		storeProperty(lifeCycle, getInitialArrayName(MAIN), dataId, properties[i], value);
		value = getInitialProperty(lifeCycle, OUTRO, dataId, properties[i]);
		storeProperty(lifeCycle, getInitialArrayName(OUTRO), dataId, properties[i], value);
	}
	
}

/**
 * Storring property value into the array with initial values for some element
 * @param {Array} array - array with life cycle
 * @param {String} arrayIndex - index of the subarray with initial values for some of the phases
 * @param {String} targetId - id of the element
 * @param {String} property - property name
 * @param {Object} value - property value
 */
function storeProperty(array, arrayIndex, targetId, property, value)
{
	if (value != null) {
		if (array[arrayIndex] == null)
			array[arrayIndex] = new Array();
		array[arrayIndex].push({
			"target": targetId,
			"property": property,
			"value": value
		});
	}
}

/**
 * Retrieves all properties for element that can be changed from life cycle
 * @param {String} elementType - type of the element of this life cycle ("frame" or "page")
 */
function getProperties(elementType)
{
	var properties = new Array();
	if (elementType == "frame")
	{
		properties.push(VISIBILITY);
		properties.push(LOCATION);
		properties.push(MEDIA_STATE);
	}
	else if (elementType == "page")
	{
		properties.push(ACTIVATION_FOR_FRAME);
	}
	return properties;
}

/**
 * Retrieves value of the last change in the current phase of the life cycle before some time
 * @param {Array} lifeCycle - life cycle array
 * @param {String} targetId - id of the element
 * @param {String} property - property name
 * @param {String} timePhase - phase in which is searching changes
 * @param {long} timeMilliseconds - time before we are searching
 */
function getPropertyValue(lifeCycle, targetId, property, timePhase, timeMilliseconds)
{
	if (timePhase == null || lifeCycle[timePhase] == null)
		return null;
	var previousChange = searchElement(lifeCycle[timePhase], 0, lifeCycle[timePhase].length - 1, timeMilliseconds);
	if (previousChange == -1)
		return null;
	for (var i=previousChange; i >= 0; i--)
	{
		if (lifeCycle[timePhase][i].property == property && lifeCycle[timePhase][i].target == targetId)
			return lifeCycle[timePhase][i].value;		
	}
	return null;
}

/**
 * Retrieves value of the property from the initial values
 * @param {Array} lifeCycle - array of the life cycle
 * @param {String} targetId - id of the element
 * @param {String} property - property name
 * @param {String} initialIndex - index of the array with the initial values for a some phase
 */
function getPropertyFromInitial(lifeCycle, targetId, property, initialIndex)
{
	if (lifeCycle[initialIndex] == null)
		return null;
	for (var i=0; i < lifeCycle[initialIndex].length; i++)
	{
		if(lifeCycle[initialIndex][i].property == property && lifeCycle[initialIndex][i].target == targetId)
			return lifeCycle[initialIndex][i].value;
	}
	return null;
}

/**
 * Searching element index of the last change before some time (sort of the binary search)
 * @param {Array} array - array of the life cycle
 * @param {int} low - low index
 * @param {int} high - high index
 * @param {long} value - time before we searching last change
 */
 function searchElement(array, low, high, value)
 {
       if (high < low)
           return -1; // not found
       mid = low + parseInt((high - low) / 2);
       if (array[mid].time > value) {
	   	if (mid > 0 && array[mid-1].time < value)
			return mid-1;
	   	return searchElement(array, low, mid - 1, value);
	   }
	   else 
	   	if (array[mid].time < value) {
			if (mid == array.length - 1) 
				return mid;
			return searchElement(array, mid + 1, high, value);
		}
		else 
			return mid; // found
 }

/**
 * Retrieves initial properties of the life cycle phase
 * @param {Array} lifeCycle - array of the life cycle
 * @param {String} phase - phase of life cycle
 * @param {String} id - id of the sophie frame, page or book
 * @param {String} property - property for retrieved info
 */
function getInitialProperty(lifeCycle, phase, id, property)
{
	return getPropertyValueInPhaseTime(lifeCycle, phase, 0, id, property);
}

/**
 * Retrieves property value in some moment from the life cycle
 * @param {Array} lifeCycle - array of the life cycle
 * @param {String} phase - phase name in which we searching value of the property
 * @param {long} milliseconds - milliseconds from the phase in which we searching value of the property
 * @param {String} id - id of the element
 * @param {String} property - property name
 */
function getPropertyValueInPhaseTime(lifeCycle, phase, milliseconds, id, property)
{
	var value = getPropertyValue(lifeCycle, id, property, phase, milliseconds);
	if (value != null)
	{
		return value;
	}
	value = getPropertyFromInitial(lifeCycle, id, property, getInitialArrayName(phase));
	if(value != null)
	{
		return value;
	}
	var previousPhase = getPreviousArrayName(phase);
	if (previousPhase == null)
		return null;
	return getPropertyValueInPhaseTime(lifeCycle, previousPhase, getLengthOfPart(lifeCycle, previousPhase), id, property);
}

/** 
 * Define map of functions which cover all possible actions that can be defines via lifeCycles
 */
var propertyDict = new Object();
propertyDict[VISIBILITY] = function(id, value) {
	//check function for existence before executing it to prevent error 
	if(document.getElementById(getCurrentPageFrameId()).contentWindow.changeVisibility != null)
		document.getElementById(getCurrentPageFrameId()).contentWindow.changeVisibility(id, value);
	/*if(value)
	{
		$("#" + id).show();
	}
	else
	{
		$("#" + id).hide();
	}*/
};
propertyDict[LOCATION] = function(id, value, pageId, elementType) {
	
	//check function for existence before executing it to prevent error 
	if(document.getElementById(getCurrentPageFrameId()).contentWindow.changeLocation != null)
		document.getElementById(getCurrentPageFrameId()).contentWindow.changeLocation(id, value);
	
}
propertyDict[ACTIVATION_FOR_FRAME] = function(id, value, pageId, elementType) {
	goTo(pageId, id, value, "frame");
}
propertyDict["pageNumber"] = function(id, value, pageId, elementType) {
	goToPage(value);
}
propertyDict[MEDIA_STATE] = function (id, value, pageId, elementType) {
	//check function for existence before executing it to prevent error 
	if(document.getElementById(getCurrentPageFrameId()).contentWindow.changeMediaState != null)
		document.getElementById(getCurrentPageFrameId()).contentWindow.changeMediaState(id, value);
}


/** 
 * Function defining basic operation in lifeCycle like properties change
 */
function changeProperty(id, property, value, pageId, elementType)
{
	if (elementType == "pageLifeCycle" && property ==  VISIBILITY)
		return;
	
	//check function for existence before evaluating it to prevent error		
	if (elementType == "pageLifeCycle" && property == ACTIVATION_FOR_FRAME &&
		document.getElementById(getCurrentPageFrameId()).contentWindow.hasGroupWithId != null &&
		document.getElementById(getCurrentPageFrameId()).contentWindow.hasGroupWithId(id)) // element is group
	{
		var elements = document.getElementById(getCurrentPageFrameId()).contentWindow.getGroupElements(id);
		for (var i = 0; i < elements.length; i++)
			changeProperty(elements[i], property, value, pageId, "frameLifeCycle");
		return;
	}
	
	propertyDict[property](id, value, pageId, elementType);
	
}

/**
 * Start life cycle of a book, page or frame
 * @param {Object} lifeCycleId - "bookLifeCycle" for a book, "pageLifeCycle" for a page or frame id for a frame
 */
function startLifeCycle(lifeCycleType, lifeCycleId)
{
	if (lifeCycleType == "bookLifeCycle" && bookTimer)
	{
		bookTimer.start();
		document.getElementById(getCurrentPageFrameId()).contentWindow.postMessage('play timeline', '*');
	}
	else if (lifeCycleType == "pageLifeCycle")
	{
		lifeCycleId = pageIds[currentPageNumber];
		if (pageTimer && pageTimer[lifeCycleId])
			pageTimer[lifeCycleId].start();
		if (frameTimers && frameTimers[lifeCycleId])
		{
			$.each(frameTimers[lifeCycleId], function(index)
			{
				if (frameTimers[lifeCycleId][index].isPaused || pageTimer[lifeCycleId] == null)
					frameTimers[lifeCycleId][index].start();
			});
		}
		//start book audio if book audio position is defined for current page
		if(bookAudioPositions != null && bookAudioPositions[lifeCycleId] != null) {
			if (pausedAudioPositions[lifeCycleId] != null) {
				//console.log("play book audio from pausedAudioPosition " + pausedAudioPositions[lifeCycleId]);
				console.log("pausedAudioPositions[lifeCycleId]" + pausedAudioPositions[lifeCycleId]);
				parent.setBookAudioPosition(pausedAudioPositions[lifeCycleId]);
			}
			else {
				//console.log("play book audio from bookAudioPosition " + bookAudioPositions[lifeCycleId]);
				console.log("bookAudioPositions[lifeCycleId]" + bookAudioPositions[lifeCycleId]);
				parent.setBookAudioPosition(bookAudioPositions[lifeCycleId]);
			}
		}
	
		
	}
	else if (frameTimers && frameTimers[lifeCycleId])
	{
		frameTimers[lifeCycleId].start();
	}
}

/**
 * Stop life cycle of a book, page or frame
 * @param {Object} lifeCycleId - "bookLifeCycle" for a book, "pageLifeCycle" for a page or frame id for a frame
 */
function stopLifeCycle(lifeCycleType, lifeCycleId)
{
	if (lifeCycleType == "bookLifeCycle" && bookTimer)
	{
		//document.getElementById(getCurrentPageFrameId()).contentWindow.postMessage('stop timeline', '*');
		//bookTimer.stop();
	}
	else if (lifeCycleType == "pageLifeCycle")
	{
		if (lifeCycleId == "this")
			lifeCycleId = pageIds[currentPageNumber];
		if (pageTimer[lifeCycleId])
			pageTimer[lifeCycleId].stop();
		if (frameTimers && frameTimers[lifeCycleId]) {
			$.each(frameTimers[lifeCycleId], function(index){
				frameTimers[lifeCycleId][index].stop();
			});
		}
		//set paused audio position of book audio for current page to undefined (null)
		pausedAudioPositions[lifeCycleId] = null;
	}
	else if (frameTimers && frameTimers[lifeCycleId])
	{
		frameTimers[lifeCycleId].stop();
	}
}

/**
 * Pause life cycle of a book, page or frame
 * @param {Object} lifeCycleId - "bookLifeCycle" for a book, "pageLifeCycle" for a page or frame id for a frame
 */
function pauseLifeCycle(lifeCycleType, lifeCycleId)
{
	if (lifeCycleType == "bookLifeCycle" && bookTimer)
	{
		bookTimer.pause();
		document.getElementById(getCurrentPageFrameId()).contentWindow.postMessage('pause timeline', '*');
	}
	else if (lifeCycleType == "pageLifeCycle")
	{
		lifeCycleId = pageIds[currentPageNumber];
		if (pageTimer[lifeCycleId])
			pageTimer[lifeCycleId].pause();
		if (frameTimers[lifeCycleId]) {
			$.each(frameTimers[lifeCycleId], function(index){
				frameTimers[lifeCycleId][index].pause();
			});
		}
		//save book audio position when life cycle is paused
		pausedAudioPositions[lifeCycleId] = parent.document.getElementById("bookAudio").currentTime;
		parent.document.getElementById("bookAudio").pause(); 
	}
	else if (frameTimers && frameTimers[lifeCycleId])
	{
		frameTimers[lifeCycleId].pause();
	}
}

/**
 * Rewind life cycle of a book, page or frame
 * @param {Object} lifeCycleId - "bookLifeCycle" for a book, "pageLifeCycle" for a page or frame id for a frame
 */
function rewindLifeCycle(lifeCycleType, lifeCycleId)
{
	if (lifeCycleType == "bookLifeCycle" && bookTimer)
	{
		bookTimer.rewind();
		document.getElementById(getCurrentPageFrameId()).contentWindow.postMessage('rewind timeline', '*');
	}
	else if (lifeCycleType == "pageLifeCycle")
	{
		lifeCycleId = pageIds[currentPageNumber];
		if (frameTimers[lifeCycleId]) {
			$.each(frameTimers[lifeCycleId], function(index)
			{
				frameTimers[lifeCycleId][index].stop();
			});
		}
		if (pageTimer)
			pagetimer.rewind();	
	}
	else if (frameTimers && frameTimers[lifeCycleIf])
	{
		frameTimers[lifeCycleId].rewind();
	}
}

/**
 * Going to a part of the life cycle of a frame
 * @param {Object} frameId - id of the sophie frame
 * @param {Object} position - position of the life cycle (phase and milliseconds)
 */
function goTo(pageId, elementId, position, elementType)
{
	var lifeCycle;
	if (elementType != "pageLifeCycle"){
		lifeCycle = frameTimers[pageId][elementId];
	}
	else {
		lifeCycle = pageTimer[pageId];
	}
	//first check wether lifeCycle exist, otherwise log exception
	if(lifeCycle == undefined)
	{
		console.log("EXCEPTION : lifeCycle is undefined for " + pageId);
		//return ;
	}
	if (position.phase != OFF) {
		lifeCycle.goToPositionOfPhase(position.phase, position.milliseconds);
	}
	else {// "off"
		lifeCycle.stop();
		propertyDict[VISIBILITY](elementId, false);
		//\$("#" + elementId).hide();
	}
}

/**
 * Class for addLifeCycle
 * @param {Object} id - id of the object which life cycle is defines
 * @param {Object} data - contains all data about life cycle of the object
 */
function addLifeCycle(id, data)
{
	//alert("create life cycle for object" + id);
	if(lifeCycles[id] == null)
		lifeCycles[id] = new Array();
		
	if(data != null)
	{
		//add event listeners for link data
		for(var i=0; i<data.length; i++)
		{
			lifeCycle[id].push({
				//"target": id,
				"property": data[i].property,
				"value": data[i].value,
				"time": data[i].time
			});
		}
	}
}

/**
 * Whether page life cycle is in outro or not
 */
function isInOutro(pageId)
{
	if (pageTimer && pageTimer[pageId] && pageTimer[pageId].positionInLifeCycle != OUTRO &&
		pageLifeCycle[pageId][OUTRO_LENGTH] != 0)
	{
		return false;
	}
	return true;
}

/**
 * Class for Timer
 * @param {Object} id - id of object which timer is running
 * @param {Object} index - parameter indicating current position in life cycle of the object with specified id
 */
function LifeCycle(id, type, lifeCyclesData, index)
{
	this.id = id;
	this.type = type;
	this.lifeCyclesData = lifeCyclesData;
	this.index = index;

	this.startTime;
	this.positionInLifeCycle = OFF; // values: "off", "intro", "main", "outro"
	this.delay = 0;
	this.pauseDelay = this.delay;
	
	this.isPaused = false;
	this.isRunning = false;
	this.timeout;
	
	this.callFunction = function(thisObj) {
		
		//time has completed, so set paused and running properties to false
		thisObj.isPaused = false;
		thisObj.isRunning = false;
		if (thisObj.index != -1)
		{ // changing properties of the life cycle when timer is finished and there is change for this time
			changeProperty(thisObj.lifeCyclesData[thisObj.positionInLifeCycle][thisObj.index].target,
						   thisObj.lifeCyclesData[thisObj.positionInLifeCycle][thisObj.index].property,
						   thisObj.lifeCyclesData[thisObj.positionInLifeCycle][thisObj.index].value, thisObj.id, thisObj.type);
		}
		else if (thisObj.positionInLifeCycle == MAIN && thisObj.lifeCyclesData[MAIN] != null)
		{ 
			if (thisObj.lifeCyclesData["looping"]) {// looping the main part
				for (var i = 0; i < thisObj.lifeCyclesData[INITIAL_MAIN].length; i++) 
					changeProperty(thisObj.lifeCyclesData[INITIAL_MAIN][i].target, thisObj.lifeCyclesData[INITIAL_MAIN][i].property, thisObj.lifeCyclesData[INITIAL_MAIN][i].value, thisObj.id, thisObj.type);
			}
			else { //go to outro
				goToNextPageAfterOutro = true;
				thisObj.goToPositionOfPhase(OUTRO, 0);
			}
		}
		if (thisObj.index > -1 && thisObj.index < thisObj.lifeCyclesData[thisObj.positionInLifeCycle].length-1) 
		{ // starting next timer
			thisObj.index = thisObj.index+1;
			thisObj.delay = thisObj.lifeCyclesData[thisObj.positionInLifeCycle][thisObj.index].time - thisObj.lifeCyclesData[thisObj.positionInLifeCycle][thisObj.index-1].time;
			thisObj.start();
		}
		else 
		{ // last change of intro, main or outro was made
			if (thisObj.positionInLifeCycle == INTRO || thisObj.positionInLifeCycle == MAIN)
			{
				if (thisObj.index == -1) { // there was no changes in the intro and intro is finished
					if (thisObj.lifeCyclesData["looping"]) // looping the main part
						thisObj.goToPositionOfPhase(MAIN, 0);
					else { //go to outro
						goToNextPageAfterOutro = true;
						thisObj.goToPositionOfPhase(OUTRO, 0);
					}
				}
				else { // last change in intro is made and the next one will be in main part of the life cycle
					thisObj.delay = getLengthOfPart(thisObj.lifeCyclesData, thisObj.positionInLifeCycle) - thisObj.lifeCyclesData[thisObj.positionInLifeCycle][thisObj.index].time;
					thisObj.index = -1;
					thisObj.start();
				}
			}
			else //thisObj.positionInLifeCycle = "outro"
			{
				if (thisObj.index != -1)
				{ // last change in outro was made, start timer for finishing the outro part of the life cycle
					thisObj.delay = thisObj.lifeCyclesData[OUTRO_LENGTH] - thisObj.lifeCyclesData[thisObj.positionInLifeCycle][thisObj.index].time;
					thisObj.index = -1;
					thisObj.start();
				}
				else
				{ // time of the outro part of the life cycle is finished: life cycle is finished
					if (thisObj.type == "pageLifeCycle") {
						if (goToNextPageAfterOutro == true) {
							goToNextPage();
							goToNextPageAfterOutro = false;
							thisObj.positionInLifeCycle = OFF;
						}
						if (goToPreviousPageAfterOutro == true) {
							goToPreviousPage();
							goToPreviousPageAfterOutro = false;
							thisObj.positionInLifeCycle = OFF;
						}
					}
					thisObj.stop();
				}
			}
		}
	}
	
	this.start = function()
	{
		if (this.positionInLifeCycle == OFF) {
			this.positionInLifeCycle = INTRO;
			if (this.lifeCyclesData[INITIAL_INTRO] != null && this.lifeCyclesData[INTRO_LENGTH] != 0)
				for (var i = 0; i < this.lifeCyclesData[INITIAL_INTRO].length; i++) 
					changeProperty(this.lifeCyclesData[INITIAL_INTRO][i].target, this.lifeCyclesData[INITIAL_INTRO][i].property, this.lifeCyclesData[INITIAL_INTRO][i].value, this.id, this.type);
			
		}
		if (this.delay == 0) {
			if (this.lifeCyclesData[this.positionInLifeCycle] != null) 
				this.delay = this.lifeCyclesData[this.positionInLifeCycle][0].time;
			else {
				this.delay = getLengthOfPart(this.lifeCyclesData, this.positionInLifeCycle);
				this.index = -1;
			}
		}
		
		if (this.delay == 0 && !this.isPaused) {
			if (getNextArrayName(this.positionInLifeCycle) != null) {
				this.goToPositionOfPhase(getNextArrayName(this.positionInLifeCycle), 0);
				return;
			}
		}
		
		if (!this.isPaused && !this.isRunning)
		{
			this.timeout = setTimeout(this.callFunction, this.delay, this);
		}
		else if (this.isPaused)
		{
			this.timeout = setTimeout(this.callFunction, this.delay - this.pauseDelay, this);
			this.isPaused = false;
		}
		var currentTime = new Date();
		this.startTime = Date.parse(currentTime.toUTCString());
		this.isRunning = true;
	}
	
	this.stop = function()
	{
		this.delay = 0;
		this.pauseDelay = this.delay;
		this.isPaused = false;
		this.index = 0;
		clearTimeout(this.timeout);
		this.isRunning = false;
		
		/*if (this.type == "pageLifeCycle")
		{*/
			this.positionInLifeCycle = OFF;
			return;
		/*}
		if (this.lifeCyclesData[INITIAL_INTRO] != null && this.lifeCyclesData[INTRO_LENGTH] != 0)
				for (var i = 0; i < this.lifeCyclesData[INITIAL_INTRO].length; i++) 
					changeProperty(this.lifeCyclesData[INITIAL_INTRO][i].target, this.lifeCyclesData[INITIAL_INTRO][i].property, this.lifeCyclesData[INITIAL_INTRO][i].value, this.id, this.type);
		else if (this.lifeCyclesData[INTRO_LENGTH] == 0)
		{
			this.goToPositionOfPhase(MAIN, 0, true);
		}
		
		this.positionInLifeCycle = OFF;*/
	}

	this.pause = function()
	{
		if (!this.isRunning)
		{
			return;
		}
		var pauseTime = new Date();
		this.pauseDelay = Date.parse(pauseTime.toUTCString()) - this.startTime;
		this.isRunning = false;
		if (this.pauseDelay > this.delay)
		{
			this.pauseDelay = this.delay;
			//return;
		}
		this.isPaused = true;
		clearTimeout(this.timeout); 
	} 
	
	this.rewind = function()
	{
		this.stop();
		this.start();
	}
	
	this.goToIntro = function()
	{
		this.rewind();
	}
	
	this.goToPositionOfPhase = function(phase, milliseconds, paused)
	{
		if (paused == null)
			paused = false;
		clearTimeout(this.timeout);
		
		this.positionInLifeCycle = phase;
		if (phase == OFF) {
			if (this.type == "pageLifeCycle") {
				if (goToNextPageAfterOutro == true) {		
					goToNextPage();
					goToNextPageAfterOutro = false;
				}
				if (goToPreviousPageAfterOutro == true) {
					goToPreviousPage();
					goToPreviousPageAfterOutro = false;
				}
			}
			this.stop();
			return;
		}
		if (!paused)
		{
			this.isRunning = false;
			this.isPaused = false;
		}
		if (lifeCyclesData[this.positionInLifeCycle] != null)
		{
			this.delay = this.lifeCyclesData[this.positionInLifeCycle][0].time;
			this.index = 0;
		}
		else
		{
			this.delay = getLengthOfPart(this.lifeCyclesData, phase);
			this.index = -1;
		}
		this.pauseDelay = this.delay;
		
		if (this.delay == 0)
		{
			if (getNextArrayName(this.positionInLifeCycle) != null) {
				this.goToPositionOfPhase(getNextArrayName(this.positionInLifeCycle), 0, paused);
				return;
			}
	
		}
		if (milliseconds == 0) {
			if (lifeCyclesData[getInitialArrayName(phase)] != null) 
				for (var i = 0; i < this.lifeCyclesData[getInitialArrayName(phase)].length; i++) 
					changeProperty(this.lifeCyclesData[getInitialArrayName(phase)][i].target,
									this.lifeCyclesData[getInitialArrayName(phase)][i].property,
									this.lifeCyclesData[getInitialArrayName(phase)][i].value,
									this.id, this.type);
		}
		else
		{
			var value;
			var dataType = "frame";
			if (this.type == "pageLifeCycle")
				dataType = "page";
			var properties = getProperties(dataType);
			for (var i=0; i < properties.length; i++)
			{
				value = getPropertyValueInPhaseTime(this.lifeCyclesData, phase, milliseconds, this.id, properties[i]);
				if (value != null)
					changeProperty(this.id,
							   	   properties[i],
							       value,
								   this.id, this.type);		
			}
		}
		
		if (!paused)
			this.start();
	}
}